#!/usr/bin/env python
# Copyright (c) 2000-2003 ActiveState SRL.
# See the file LICENSE.txt for licensing information.

import sys, os, time, threading
import socket
from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject

import logging

import process, koprocessutils
import uriparse

log = logging.getLogger("guibuilder")
# log.setLevel(logging.DEBUG)

class SocketWrapper:
    " Class wrapper around a socket to provide .readline() behavior "
    def __init__(self, socket):
        self._socket = socket
        self._lineBuf = ''
        
    def readline(self):
        if self._socket is not None:
            line = ''
            while 1:
                #XXX This is not portable to the Mac.
                idx = self._lineBuf.find('\n')
                if idx != -1:
                    line, self._lineBuf =\
                        self._lineBuf[:idx+1], self._lineBuf[idx+1:]
                    break
                else:
                    lengthBefore = len(self._lineBuf)
                    try:
                        t = self._socket.recv(4092)
                    except socket.error:
                        # the connection was closed.
                        return ''
                    if len(t) <= lengthBefore: # no new data was read
                        line, self._lineBuf = self._lineBuf, ""
                        break
                    else:
                        self._lineBuf += t
            return line
        else:
            raise "SocketWrapper.readline: no socket to read from"

class Listener:
    def __init__(self, observerSvc, mgr, fileurl, socket):
        self.mgr = mgr
        self.fileurl = fileurl
        self.pipe = SocketWrapper(socket)
        self.observerSvc = observerSvc
        
    def __call__(self):
        laststuff = ''
        if self.observerSvc is None:
            self._getObserver()
        while 1:
            line = self.pipe.readline()
            line = line.strip()
            if not line: break
            log.debug("Got line: %s", line)
            if line == 'exit': continue
            try:
                command, line = line.split(' ', 1)
            except ValueError:
                log.error("Error parsing line %r", line)
                continue
            if command == 'exit':
                break
            elif command == 'ping':
                # XXX: This should not be necessary, but on Win2K only the
                # guibuilder can hang when not getting data down the stdin
                # pipe - so this exists to ping/pong
                self.mgr.notifyPong(self.fileurl)
                continue
            elif command == 'interpreter':
                language = line
                self.mgr.notifyInterpreter(self.fileurl, language)
                continue
            elif command == 'help':
                if sys.platform.startswith('win'):
                    line = line.replace('/', '\\')
                # print "help: line = " + line
                url = uriparse.localPathToURI(line);
                data = '%s\t%s' % (self.fileurl, url)
                self.observerSvc.notifyObservers(None, 'guiBuilder-help', data)
                continue
            try:
                project, line = line.split(' ', 1)
            except:
                # print "Unknown GUI command 2", line 
                continue # unknown command
            if command == 'save':
                if sys.platform.startswith('win'):
                    line = line.replace('/', '\\')
                log.debug("save: line = %s", line)
                url = uriparse.localPathToURI(line);
                data = '%s\t%s' % (self.fileurl, url)
                #print line +'\n' + url +'\n'+ data
                self.observerSvc.notifyObservers(None, 'guiBuilder-save', data)
            elif command == 'startup':
                if sys.platform.startswith('win'):
                    line = line.replace('/', '\\')
                #print "startup: line = " + line
                url = uriparse.localPathToURI(line);
                data = '%s\t%s' % (self.fileurl, url)
                #print line +'\n' + url +'\n'+ data
                self.observerSvc.notifyObservers(None, 'guiBuilder-startup', data)
                continue
            elif command == 'edit':
                if sys.platform.startswith('win'):
                    line = line.replace('/', '\\')
                #print "edit: line = " + line
                url = uriparse.localPathToURI(line);
                data = '%s\t%s' % (self.fileurl, url)
                self.observerSvc.notifyObservers(None, 'guiBuilder-edit', data)
                continue
            elif command == 'dirty':
                dirtyflag = line.strip()
                if dirtyflag == '1':
                    isdirty = 1
                else:
                    isdirty = 0
                self.mgr.notifyDirty(self.fileurl, isdirty)
                continue
            else:
                #print "Unknown GUI command", line
                pass
        # if we get here it's because the pipe was closed.  Need to tell the original thread...
        self.mgr.forget(self.fileurl)

    def close(self):
        log.debug("Gui builder closed pipe");

lastErrorSvc = None

class KoGuiBuilderService:
    _com_interfaces_ = [components.interfaces.koIGuiBuilderService]
    _reg_clsid_ = "8EE42AA0-AD35-406d-8F57-9EB71622356A"
    _reg_contractid_ = "@activestate.com/koGuiBuilderService;1"
    _reg_desc_ = "Service for interacting with the GUI Builder"

    def __init__(self):
        self._socketMap = {}
        self._dirtyURLs = {}
        self._lock = threading.Lock()
        self._connectEvents = {}
        self._languages = None
        self._socket = None
        self.infoSvc = None
        self.sysUtilsSvc = None
        global lastErrorSvc
        lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)

        proxyManager = components.classes["@mozilla.org/xpcomproxy;1"].\
                            getService(components.interfaces.nsIProxyObjectManager)
        observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)

        self.observerSvc = getProxyForObject(None, components.interfaces.nsIObserverService,
                                             observerSvc, PROXY_ALWAYS | PROXY_ASYNC)
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
                            getService(components.interfaces.koIPrefService).prefs


    def getLanguages(self):
        if self._languages is not None:
            return self._languages
        guibuilder = self._getGuibuilderPath()
        argv = [guibuilder, "--interactive", "--languages"]
        env = koprocessutils.getUserEnv()
        p = process.ProcessOpen(argv, env=env)
        s = p.stdout.read()
        p.close()
        s = s.strip()
        languages = s.split('\t')
        ids = []
        labels = []
        for language in languages:
            id, label = language.split(':')
            ids.append(id)
            labels.append(label)
        self._languages = ids, labels
        return ids, labels

    def getDirtyURLs(self):
        return self._dirtyURLs.keys()

    def getOpenURLs(self):
        return self._socketMap.keys()
    
    def notifyDirty(self, url, dirtyflag):
        # this is called from other threads, so needs guarding.
        self._lock.acquire()
        if dirtyflag:
            self._dirtyURLs[url] = dirtyflag
        else:
            if url in self._dirtyURLs:
                del self._dirtyURLs[url]
        self._lock.release()

    def notifyPong(self, url):
        # XXX: This should not be necessary, but on Win2K only the
        # guibuilder can hang when not getting data down the stdin
        # pipe - so this exists to ping/pong
        self._socketMap[url].sendall('pong\n')

    def notifyInterpreter(self, url, language):
        if self.infoSvc == None:
            self.infoSvc = components.\
                           classes["@activestate.com/koInfoService;1"].\
                           getService(components.interfaces.koIInfoService)
        if self.sysUtilsSvc == None:
            self.sysUtilsSvc = components.\
                               classes["@activestate.com/koSysUtils;1"].\
                               getService(components.interfaces.koISysUtils)
        if language == 'perl':
            path = self._globalPrefs.getStringPref("perlDefaultInterpreter");
            if path == '':
                if self.infoSvc.isWindows:
                    path = 'wperl'
                else:
                    path = 'perl'
                path = self.sysUtilsSvc.Which(path)
                # If there is none, then GUI builder will deal with it.
        elif language == 'tcl':
            path = self._globalPrefs.getStringPref("wishDefaultInterpreter")
            if path == '':
                path = self.sysUtilsSvc.Which('wish')
        elif language == 'tkinter':
            path = self._globalPrefs.getStringPref("pythonDefaultInterpreter")
            if path == '':
                if self.infoSvc.isWindows:
                    path = 'pythonw'
                else:
                    path = 'python'
                path = self.sysUtilsSvc.Which(path)
        else:
            log.error("Unknown language received in notifyInterpreter: %r", language)
            return
            self._socketMap[url].sendall('\n') # assuming the gui builder wants an answer
        if self.infoSvc.isWindows:
            path = path.replace('\\', '/')
        self._socketMap[url].sendall('interpreter %s %s\n' % (language, path))

    def forget(self, fileurl):
        log.debug("Forgetting %s", fileurl)
        filepath = uriparse.URIToLocalPath(fileurl)
        if not filepath or not os.path.exists(filepath):
            self.observerSvc.notifyObservers(None, 'guiBuilder-forget', fileurl)
        # this is called from other threads, so needs guarding.
        self._lock.acquire()
        if fileurl in self._socketMap:
            self._socketMap[fileurl].close()
            del self._socketMap[fileurl]
        if fileurl in self._dirtyURLs:
            del self._dirtyURLs[fileurl]
        if fileurl in self._connectEvents:
            del self._connectEvents[fileurl]
        self._lock.release()

    def _getGuibuilderPath(self):
        """Return the full path the GUI Builder executable.
        
        If the GUI Builder exe could not be found a suitable execption
        is raised.
        """
        dirSvc = components.classes["@activestate.com/koDirs;1"]\
                 .getService(components.interfaces.koIDirs)
        dname = os.path.join(dirSvc.installDir, "guibuilder")
        if sys.platform.startswith("win"):
            fname = os.path.join(dname, "guibuilder.exe")
        else:
            fname = os.path.join(dname, "guibuilder")
        if os.path.exists(fname):
            return fname
        else:
            ex = "Could not find the GUI Builder executable."
            lastErrorSvc.setLastError(0, ex)
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, ex)

    def _connect(self, fileurl, language=None):
        guibuilder = self._getGuibuilderPath()
        filepath = uriparse.URIToLocalPath(fileurl)
        dname, fname = os.path.split(filepath)
        dname = dname.replace('\\', '/')

        port = self._globalPrefs.getLongPref('gui_port')

        argv = [guibuilder, "--interactive", str(port), "--directory", dname]

        # Always provide some language argument to avoid the guibuilder's
        # query language dialog
        if language:
            argv.extend(['--language', language])
        else:
            argv.extend(['--language', 'default'])

        if not self._socket:
            self._socket = self._make_socket(port)

        log.debug("Connecting to %s", argv)
        env = koprocessutils.getUserEnv()
        p = process.Process(argv, env=env)

        conn, (remotehost, remoteport) = self._socket.accept()

        threading.Thread(target=Listener(self.observerSvc, self, fileurl, conn)).start()
        self._socketMap[fileurl] = conn
        return p, conn, fname

    def _make_socket(self, port):
        log.debug("in _make_socket: %s", port)
        # XXX need pref observer to pick up on changes in the port, that would close the port 
        # and restart it.
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(('127.0.0.1', port))
        s.listen(10)
        return s

    # Edit and Create are async calls -- since process creation can be slow
    # and can hang -- we don't want to hang the main thread.
    # We'll take the chance for save/close

    def Open(self, fileurl, language):
        """Start a connection with a new (or existing) GUI builder
        viz. a particular _existing_ project"""
        log.debug("Doing open for %s, %s", fileurl, language)
        if fileurl in self._socketMap:
            log.debug("found URL in socketMap")
            i = self._socketMap[fileurl]
            try:
                log.debug("sending deiconify")
                i.sendall('wm deiconify .\n')
                log.debug("sent")
                return 1 # we're done.
            except OSError, e:
                # if this happens it means that the gui builder quit
                # without telling us about it.  We'll clean up our
                # data structures ourselves and
                # restart it.
                log.exception("Got an error doing the deiconify")
                self.forget(fileurl) # fall through to the non-existent connection case.
        self._connectEvents[fileurl] = threading.Event()
        t = threading.Thread(target=self._connectAndSendAll, args=("load", fileurl, language))
        t.start()
        return 1
    
    def Create(self, fileurl, language):
        """Start a connection with a new (or existing) GUI builder
        viz. a particular _non-existing_ project"""
        self._connectEvents[fileurl] = threading.Event()
        t = threading.Thread(target=self._connectAndSendAll, args=("new", fileurl, language))
        t.start()
        return 1
    
    def _connectAndSendAll(self, command, fileurl, language):
        try:
            p, conn, fname = self._connect(fileurl, language)
            # Send the language as well.  Really only need for 'new',
            # but is just ignored for 'load'.
            conn.sendall(command + ' ' + fname + ' "' + language + '"\n')
        except OSError, ex:
            self.forget(fileurl)
            log.error("Exception in _connectAndSendAll: %r", ex)
            msg = "Couldn't tell the GUI builder to load " + fname
            lastErrorSvc.setLastError(0, msg)
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, msg)
        self._connectEvents[fileurl].set()

    def Save(self, fileurl):
        assert fileurl in self._socketMap
        try:
            self._connectEvents[fileurl].wait()
            self._socketMap[fileurl].sendall('save\n');
        except OSError, ex:
            self.forget(fileurl)
            log.error("Exception in Save: %r", ex)
            msg = "Couldn't tell the GUI builder to save"
            lastErrorSvc.setLastError(0, msg)
            raise ServerException(nsError.NS_ERROR_INVALID_ARG, msg)
        return 1
        
    def Quit(self, fileurl):
        assert fileurl in self._socketMap
        self._connectEvents[fileurl].wait()
        try:
            self._socketMap[fileurl].sendall('dirty 0\n');
            self._socketMap[fileurl].sendall('exit\n');
        except:
            pass
        self.forget(fileurl)
        return 1

    def QuitAll(self):
        filepaths = self._socketMap.keys()
        for filepath in filepaths:
            if not self.Quit(filepath):
                return 0
        return 1
